"""Module for interaction with GIT."""
import glob
import logging
import os
import shutil
import subprocess
from dataclasses import dataclass
from enum import Enum
from typing import Iterator, List, Optional


class GITError(Exception):
    """Any error related with GIT usage."""


class CommandError(Exception):
    """Command error."""


class ProjectStatus(Enum):
    """GIT project status."""

    CLEAN = "clean"
    DIRTY = "dirty"
    UNDEFINED = "undefined"


@dataclass
class ProjectInfo:
    """GIT Project information."""

    name: str
    status: Optional[ProjectStatus]


def _run_command(
    command: str, check=False, cwd: str = None
) -> subprocess.CompletedProcess:
    """Run command in subprocess."""
    logging.debug('Running command "%s"', command)
    return subprocess.run(
        command.split(), cwd=cwd, capture_output=True, text=True, check=check
    )


def is_git_dir(directory: str) -> bool:
    """Return whether a directory is GIT initialized directory."""
    return os.path.isdir(directory) and ".git" in os.listdir(directory)


def _get_stash_info(directory: str):
    """Return stash info under `directory`."""
    logging.debug('Checking for unpushed GIT stashes under "%s"', directory)
    return _run_command("git stash list", cwd=directory).stdout


def _get_unpushed_branches_info(directory: str) -> str:
    """Return information about unpushed branches.

    Format is: <commit> (<branch>) <commit_message>
    """
    logging.debug('Checking for unpushed GIT commits under "%s"', directory)
    return _run_command(
        "git log --branches --not --remotes --decorate --oneline", cwd=directory
    ).stdout


def _get_unstaged_info(directory: str) -> str:
    """Return information about unstaged changes."""
    logging.debug('Checking for unstaged changes under "%s"', directory)
    return _run_command("git status --short", cwd=directory).stdout


def _get_unpushed_tags(directory: str) -> str:
    """Return unpushed tags.

    If no tags found, returns an empty string.
    If failed to get tags information, returns a string containing error
    description.
    """
    logging.debug('Checking for unpushed tags under "%s"', directory)

    try:
        info = _run_command(
            "git push --tags --dry-run", cwd=directory, check=True
        ).stderr
    except subprocess.CalledProcessError as exc:
        return f"Failed to check unpushed tags: {exc.stderr}"

    if "new tag" not in info:
        return ""
    return info


def check_all_pushed(directory: str) -> None:
    """Check if everything from GIT directory is pushed.

    It checks:
      * stashes
      * branches
      * unstaged
      * tags

    :raises: `GITError` if there is something unpushed. Error message contains
      information about unpushed entities
    """
    unstaged = _get_unstaged_info(directory)
    stashes = _get_stash_info(directory)
    branches = _get_unpushed_branches_info(directory)
    tags = _get_unpushed_tags(directory)

    if any([unstaged, stashes, branches, tags]):
        output = ""
        if stashes:
            output += f"Stashes:\n{stashes}"
        if branches:
            output += f"\nCommits:\n{branches}"
        if unstaged:
            output += f"\nNot staged:\n{unstaged}"
        if tags:
            output += f"\nTags:\n{tags}"

        raise GITError(output)


def clone(source: str, destination: str):
    """Clone a project from GIT `source` to `destination` directory."""
    try:
        logging.info('Cloning "%s" to "%s"', source, destination)
        _run_command(f"git clone {source} {destination}", check=True)
    except subprocess.CalledProcessError as exc:
        raise GITError(f'Failed to clone "{source}":\n{exc.stderr}') from exc


class WorkingDir:
    """Encapsulates working directory for GIT projects."""

    def __init__(self, directory: str) -> None:
        self.directory = os.path.expanduser(directory)
        self._ensure_directory()

    def _ensure_directory(self) -> None:
        os.makedirs(self.directory, exist_ok=True)

    @property
    def _dirs(self) -> List[str]:
        return os.listdir(self.directory)

    def remove(self, project_name: str = None, force: bool = False) -> None:
        """Remove project from the directory.

        If `project_name` is not specified, all projects will be removed.
        """
        if project_name:
            if project_name not in self._dirs:
                raise CommandError(f'"{project_name}" not found in "{self.directory}"')
            self._remove_project(project_name, force)
        else:
            self._remove_projects(force)

    def clone(self, project_name: str, sources: List[str]) -> None:
        """Clone a project to the working directory."""
        if project_name in self._dirs:
            raise CommandError(f'Project "{project_name}" is already cloned')

        for i, source in enumerate(sources, start=1):
            try:
                clone(
                    os.path.join(source.strip("/"), f"{project_name}.git"),
                    f"{self.directory}/{project_name}",
                )
                break
            except GITError as exc:
                if i == len(sources):
                    raise CommandError(
                        f'Failed to clone "{project_name}". Tried all configured sources'
                    ) from exc
                logging.debug(exc)

    def open(self, project_name: str, editor: str = None) -> None:
        """Open a project from the directory.

        If editor is not specified, try $EDITOR, vi, vim consequently.
        """
        project_dir = os.path.join(self.directory, project_name)

        if not os.path.isdir(project_dir):
            raise CommandError(
                f'No project named "{project_name}" found under the working directory'
            )

        for editor_ in (editor, os.environ.get("EDITOR"), "vi", "vim"):
            if editor_:
                logging.info('Opening "%s" with "%s" editor', project_dir, editor_)
                result = subprocess.run([editor_, project_dir], check=False)
                if result.returncode == 0:
                    break
        else:
            raise CommandError(f'No suitable editor found to open "{project_dir}"')

    def show(self, check_status: bool) -> Iterator[ProjectInfo]:
        """Return information about GIT projects."""
        for project in self._dirs:
            yield ProjectInfo(
                project, self._get_project_status(project) if check_status else None
            )

    def _get_project_status(self, project_name: str) -> ProjectStatus:
        path = os.path.join(self.directory, project_name)
        if not is_git_dir(path):
            return ProjectStatus.UNDEFINED
        try:
            check_all_pushed(path)
        except GITError:
            return ProjectStatus.DIRTY
        else:
            return ProjectStatus.CLEAN

    def _remove_projects(self, force: bool = False) -> None:
        for project in self._dirs:
            if os.path.isdir(os.path.join(self.directory, project)):
                try:
                    self._remove_project(project, force=force)
                except CommandError as exc:
                    logging.error(exc)
                    continue
        # there may be some files left
        for filepath in glob.glob(os.path.join(self.directory, "*")):
            if os.path.islink(filepath):
                logging.debug('Removing symlink "%s"', filepath)
                os.unlink(filepath)
            elif not os.path.isdir(filepath):
                logging.debug('Removing file "%s"', filepath)
                os.remove(filepath)

    def _remove_project(self, project_name: str, force: bool = False) -> None:
        logging.info('Finishing up "%s"', project_name)
        proj_path = os.path.join(self.directory, project_name)

        if not is_git_dir(proj_path):
            logging.debug('Not a GIT repository, removing "%s"', proj_path)
            shutil.rmtree(proj_path)
            return

        try:
            if force or check_all_pushed(proj_path) is None:
                logging.debug('Removing "%s"', proj_path)
                shutil.rmtree(proj_path)
        except GITError as exc:
            raise CommandError(
                f"There are some unpushed changes or problems! See below\n\n"
                f"{exc}\n"
                f'Push your local changes or use "-f" flag to drop them'
            ) from exc

    def __contains__(self, item) -> bool:
        return item in self._dirs



"""Module for dealing with the utility configuration."""
import json
import logging
import os
from dataclasses import dataclass
from typing import Optional

import appdirs

_CONFIG_PATH = os.path.join(appdirs.user_config_dir("git_workon"), "config.json")
_CONFIG_TEMPLATE = {
    "dir": "~/git_workon",
    "editor": "",
    "sources": [],
}


class ConfigError(Exception):
    """Configuration error."""


@dataclass
class UserConfig:
    """User configuration."""

    dir: Optional[str]
    editor: Optional[str]
    sources: Optional[list]

    def __post_init__(self):
        if self.dir and not isinstance(self.dir, str):
            raise ConfigError('"dir" parameter should be of string type')
        if self.editor and not isinstance(self.editor, str):
            raise ConfigError('"editor" parameter should be of string type')
        if self.sources and not isinstance(self.sources, list):
            raise ConfigError('"sources" parameter should be of array type')


def load_config(path: str = _CONFIG_PATH) -> "UserConfig":
    """Load a configuration from path."""
    try:
        with open(path, encoding="utf8") as file:
            config = json.load(file)
    except (json.JSONDecodeError, OSError):
        config = {}

    return UserConfig(config.get("dir"), config.get("editor"), config.get("sources"))


def init_config(path: str = _CONFIG_PATH) -> None:
    """Initialize configuration.

    If the config file does not exists -> create it.
    """
    logging.info("Configuration is under %s", path)
    os.makedirs(os.path.dirname(path), exist_ok=True)

    if not os.path.exists(path):
        logging.debug('Writing configuration template to "%s"', path)
        with open(path, "w", encoding="utf8") as file:
            json.dump(_CONFIG_TEMPLATE, file, indent=2)



"""Command Line Interface for the GIT workon."""
import argparse
import logging
import sys
from dataclasses import dataclass
from typing import Iterator, List, Optional

import termcolor

from . import config as config_module
from . import git


class CLIError(Exception):
    """CLI error."""


# pylint:disable=too-few-public-methods
class ExtendAction(argparse.Action):
    """Extend action for `argparse`."""

    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest) or []
        items.extend(values)
        setattr(namespace, self.dest, items)


_COLOR_FOR_STATUS = {
    git.ProjectStatus.CLEAN: "green",
    git.ProjectStatus.DIRTY: "yellow",
    git.ProjectStatus.UNDEFINED: "white",
}


@dataclass
class ArgParseArgument:
    """Wrapper encapsulating `argparse` argument."""

    positional: tuple
    keyword: dict


def _append_args(parser, args: Optional[List[ArgParseArgument]]) -> None:
    if args:
        for arg in args:
            parser.add_argument(*arg.positional, **arg.keyword)


def _append_start_command(subparsers, parent, user_config: config_module.UserConfig):
    start_parser = subparsers.add_parser(
        "start",
        help="start your work on a project",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[parent],
        add_help=False,
    )
    start_parser.register("action", "extend", ExtendAction)

    start_parser.add_argument("project", help="project name to start with")
    start_parser.add_argument(
        "-s",
        "--source",
        help="git source including username",
        action="extend",
        nargs="+",
        required=user_config.sources is None,
    )
    start_parser.add_argument(
        "-n",
        "--no-open",
        dest="noopen",
        help="don't open a project",
        action="store_true",
    )

    return start_parser


def _append_done_command(subparsers, parent):
    done_parser = subparsers.add_parser(
        "done",
        help="finish your work and clean working directory",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[parent],
        add_help=False,
    )

    done_parser.add_argument(
        "project",
        nargs="?",
        help=(
            "project name to finish work for. If not "
            "specified, all projects will be finished"
        ),
    )
    done_parser.add_argument(
        "-f",
        "--force",
        help=(
            "force a project directory removal even if "
            "there are some unpushed/unstaged changes or stashes"
        ),
        action="store_true",
    )

    return done_parser


def _append_config_command(subparsers, parent):
    return subparsers.add_parser(
        "config",
        help="init/show configuration",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[parent],
        add_help=False,
    )


def _append_show_command(subparsers, parent):
    show_parser = subparsers.add_parser(
        "show",
        help="list projects under the working directory",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[parent],
        add_help=False,
    )
    show_parser.add_argument(
        "-n",
        "--no-check",
        dest="nocheck",
        help="don't check projects status",
        action="store_true",
    )
    return show_parser


def _parse_args(user_config: config_module.UserConfig):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        dest="command",
        title="script commands",
        help="command to execute",
        required=True,
    )
    parent_parser = argparse.ArgumentParser()
    parent_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="get more information of what's going on",
    )

    directory_arg = ArgParseArgument(
        positional=("-d", "--directory"),
        keyword={
            "help": "working directory",
            "default": user_config.dir,
            "required": user_config.dir is None,
        },
    )
    editor_arg = ArgParseArgument(
        positional=("-e", "--editor"),
        keyword={
            "help": "editor used to open a project/configuration",
            "default": user_config.editor,
        },
    )

    start_parser = _append_start_command(subparsers, parent_parser, user_config)
    done_parser = _append_done_command(subparsers, parent_parser)
    _append_config_command(subparsers, parent_parser)
    show_parser = _append_show_command(subparsers, parent_parser)

    _append_args(
        start_parser,
        [
            directory_arg,
            editor_arg,
        ],
    )
    _append_args(
        done_parser,
        [
            directory_arg,
        ],
    )
    _append_args(
        show_parser,
        [
            directory_arg,
        ],
    )

    return parser.parse_args()


def _init_logger(verbose):
    level = logging.DEBUG if verbose >= 1 else logging.INFO
    logging.basicConfig(level=level, format="%(message)s")


def main():
    """Execute the script commands."""
    try:
        user_config = config_module.load_config()
        args = _parse_args(user_config)
        _init_logger(args.verbose)

        FUNC_FOR_COMMAND[args.command](args, user_config)
    except KeyboardInterrupt:
        logging.info("\nCanceled by user")
        sys.exit(0)
    except config_module.ConfigError as exc:
        logging.error("Configuration error: %s", exc)
        sys.exit(1)
    except git.CommandError as exc:
        logging.error("Command error: %s", exc)
        sys.exit(1)
    except Exception as exc:  # pylint:disable=broad-except
        logging.error("Unexpected script error: %s", exc)
        sys.exit(2)


def handle_start_command(
    args: argparse.Namespace,
    user_config: config_module.UserConfig,
) -> None:
    """Process start command."""
    workon_dir = git.WorkingDir(args.directory)

    if user_config.sources:
        if args.source:
            args.source.extend(user_config.sources)
        else:
            args.source = user_config.sources

    if args.project not in workon_dir:
        workon_dir.clone(args.project, args.source)

    if not args.noopen:
        workon_dir.open(args.project, args.editor)


# pylint:disable=unused-argument
def handle_done_command(
    args: argparse.Namespace,
    user_config: config_module.UserConfig,
) -> None:
    """Process done command."""
    workon_dir = git.WorkingDir(args.directory)

    if args.project:
        args.project = args.project.strip("/ ")

    workon_dir.remove(args.project, args.force)


def handle_config_command(
    args: argparse.Namespace,
    user_config: config_module.UserConfig,
) -> None:
    """Process config command."""
    config_module.init_config()
    logging.info(config_module.load_config())


def _build_projects_info_text(projects_info: Iterator[git.ProjectInfo]) -> str:
    return "\n".join(
        termcolor.colored(
            f"{info.name}",
            _COLOR_FOR_STATUS[info.status or git.ProjectStatus.UNDEFINED],
        )
        for info in projects_info
    )


def handle_show_command(
    args: argparse.Namespace,
    user_config: config_module.UserConfig,
) -> None:
    """Process show command."""
    workon_dir = git.WorkingDir(args.directory)
    projects_info = workon_dir.show(check_status=not args.nocheck)
    logging.info(_build_projects_info_text(projects_info))


# pylint:enable=unused-argument


FUNC_FOR_COMMAND = {
    "start": handle_start_command,
    "done": handle_done_command,
    "config": handle_config_command,
    "show": handle_show_command,
}

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['git_workon']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0', 'termcolor>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['gw = git_workon.cli:main']}

setup_kwargs = {
    'name': 'git-workon',
    'version': '2.1.0',
    'description': 'Utility that automates projects clone/remove and checks for nonpushed changes on removal',
    'long_description': '[![tests](https://github.com/ReturnedVoid/workon/actions/workflows/ci.yml/badge.svg)](https://github.com/ReturnedVoid/workon)\n\n# GIT workon\n\nDo you often need to clone some project, solve one task and remove it from your filesystem?\n\nAre you often afraid that you might leave something unpushed or stashed?\n\nDo you like to leave a perfectly clean desk after your work is done?\n\nThen this script is for you.\n\n## Installation\n\nThe package is available on `PyPi` and can be installed with `pip`:\n\n```bash\npip install git-workon\n```\n\n## Usage\n\n### Configuration\n\nThe script commands can be fully controlled by CLI arguments, but it is much convenient to have a configuration file\ndefining most of parameters. There is a special `config` command that will help you to prepare suitable configuration.\n\n```bash\ngw config [options]\n```\n\nThis command will:\n\n* Create configuration directory if it does not exist. It will use OS-specific config directory, e.g.\n  `~/.config/git_workon` for Linux\n* Copy template configuration file to the configuration directory if it does not exist\n\nThe configuration file is a simple JSON contains the following parameters:\n\n* `sources` - the array of sources from which projects will be cloned. Clone attempts will be done sequentially.\n  Example:\n\n  ```json\n  "sources": [\n    "https://github.com/<my_username>",\n    "git@github.com:<my_username>"\n  ]\n  ```\n\n  May be overridden by `-s/--source` argument. You can also define multiple sources: `-s first second -s third`\n* `dir` - the working directory. All projects will be cloned to this directory. May be overridden by `-d/--directory`\n  argument. `~` in path is supported\n* `editor` - the editor used to open a cloned project or the configuration. May be overridden by `-e/--editor` argument.\n  If not specified and `-e/--editor` argument is not provided, the script will try to use the editor specified by\n  `$EDITOR` environment variable. If that variable is not set, the script will try `vi` and `vim` consequently\n\nConfiguration example:\n\n```json\n{\n  "dir": "~/git_workon",\n  "editor": "vim",\n  "sources": [\n    "https://github.com/pallets",\n    "https://github.com/pypa"\n  ]\n}\n```\n\n### Start to work on a project\n\nWhen it is time to work on some project, use the `start` command:\n\n```bash\ngw start <my_project> [options]\n```\n\nThis command will:\n\n* If the project with a given name already exists in the working directory:\n  * open it with a configured editor\n* If the project with a given name does not exist:\n  * clone it from git sources into the working directory\n  * open the project with a configured editor\n\nSee `gw start --help` for other available options on how to control the command.\n\n### Finish your work with a project\n\nWhen you are done with your work, use `done` command:\n\n```bash\ngw done [<my_project>] [options]\n```\n\nThis command will:\n\n* Check for left stashes\n* Check for unpushed commits\n* Check for left unstaged changes\n* Check for unpushed tags\n* If anything from above was not pushed:\n  * fail with an error describing what was left unpushed\n* If everything was pushed:\n  * remove a project from the working directory\n\nIf a project name was not passed, the command will try to remove everything from the working directory (still,\nfolders containing ".git" will be checked for unpushed entities).\n\nSee `gw done --help` for other available options on how to control the command.\n\n### Show all tracked projects\n\nTo list all projects under the working directory, use `show` command:\n\n```bash\ngw show [options]\n```\n\nThis command will check every project status and colorize the output according to it:\n\n* Clean (everything is pushed) - green color\n* Dirty (something is not pushed) - yellow color\n* Undefined (not a git project) - white color\n\nSee `gw show --help` for other available options on how to control the command.\n\n## Bash completions\n\nImplemented as a bash script `workon_completions`. Currently, it adds completions only for basic commands.\nTo enable completions, simply copy the script to `/etc/bash_completion.d/` or copy it anywhere and source when you\nneed.\n',
    'author': 'Andrey Nechaev',
    'author_email': 'andrewnech@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ReturnedVoid/workon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)
